#
# BTZen - library to asynchronously access Bluetooth devices.
#
# Copyright (C) 2015-2022 by Artur Wroblewski <wrobell@riseup.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import dataclasses as dtc
import typing as tp
from functools import singledispatch

from ..data import T, AddressType, Trigger, TriggerCondition
from ..device import DeviceBase, Device, DeviceTrigger
from ..service import S, Service, ServiceCharacteristic

import pytest

Tf = tp.TypeVar('Tf', bound=float)

@dtc.dataclass(frozen=True)
class ServiceEnvSensing(ServiceCharacteristic):
    pass

@singledispatch
def read(dev: DeviceBase[Service, Tf]) -> Tf:
    return 1.0  # type: ignore

@read.register
def _read_service(dev: Device[Service, Tf]) -> Tf:
    return dev.convert(b'11')

@read.register
def _read_service_chr(dev: Device[ServiceCharacteristic, Tf]) -> Tf:
    return dev.convert(b'22')

@read.register
def _read_service_chr_tr(dev: DeviceTrigger[ServiceCharacteristic, Tf]) -> Tf:
    return dev.convert(b'33')

def to_int(v: bytes) -> int:
    return int(v.decode())

INSTANCE_DATA = [
    [
        Device(Service('aaa'), '1', AddressType.PUBLIC, to_int),
        Device[Service, int],
        _read_service,
        11,
    ],
    [
        Device(ServiceCharacteristic('aaa', 'bbb', 4), '1', AddressType.PUBLIC, str),
        Device[ServiceCharacteristic, int],
        _read_service_chr,
        "b'22'",
    ],
    [
        DeviceTrigger(
            ServiceCharacteristic('aaa', 'bbb', 4),
            '1',
            AddressType.PUBLIC,
            str,
            Trigger(TriggerCondition.FIXED_TIME, 1),
        ),
        DeviceTrigger[ServiceCharacteristic, int],
        _read_service_chr_tr,
        "b'33'",
    ],
]

SUBCLASS_DATA = [
    [
        Device[Service, int],
        Device[ServiceCharacteristic, int],
    ],
    [
        Device[Service, int],
        Device[ServiceEnvSensing, int],
    ],
    [
        Device[ServiceCharacteristic, int],
        Device[ServiceEnvSensing, int],
    ],
]

NON_SUBCLASS_DATA = [
    [
        Device[ServiceCharacteristic, int],
        DeviceTrigger[ServiceCharacteristic, int],
    ],
    [
        Device[Service, int],
        DeviceTrigger[ServiceCharacteristic, int],
    ],
]

@pytest.mark.parametrize('obj, cls, func, func_result', INSTANCE_DATA)
def test_instance(obj, cls, func, func_result) -> None:  # type: ignore
    """
    Test instance of parametrised class.
    """
    assert isinstance(obj, cls)
    assert type(obj) == cls
    assert dtc.is_dataclass(obj)

@pytest.mark.parametrize('parent, cls', SUBCLASS_DATA)
def test_subclass(parent: type, cls: type) -> None:
    """
    Test inheritance of parametrised subclasses.
    """
    assert parent in cls.mro()
    assert issubclass(cls, parent)

@pytest.mark.parametrize('parent, cls', NON_SUBCLASS_DATA)
def test_not_subclass(parent: type, cls: type) -> None:
    """
    Test lack of inheritance of parametrised subclasses.
    """
    assert parent not in cls.mro()
    assert not issubclass(cls, parent)

@pytest.mark.parametrize('obj, cls, func, func_result', INSTANCE_DATA)
def test_dispatch(obj, cls, func, func_result) -> None:  # type: ignore
    """
    Test generic functions dispatch against instances of parametrised
    classes.
    """
    assert read.dispatch(type(obj)) == func
    assert read(obj) == func_result

def test_cls_create_self() -> None:
    """
    Test creating parametrised class when itself has to be returned.
    """
    t = Device[S, T]
    assert t == Device

@pytest.mark.parametrize('cls', (T, int, tp.Any))
def test_cls_create(cls: tp.Any) -> None:
    """
    Test creating parametrised class.
    """
    t = Device[Service, T]
    assert isinstance(t, type)

@pytest.mark.parametrize('cls_param', [(Service,), (Service, 'a value')])
def test_cls_param_invalid(cls_param: type) -> None:
    """
    Test if error is raised on an invalid parameters passed to a class.
    """
    with pytest.raises(TypeError) as ctx:
        Device[cls_param]   # type: ignore

    expected = 'Class subscript requires dataclass and type variable. Got {}' \
        .format(cls_param)
    assert str(ctx.value) == expected

# vim: sw=4:et:ai



#
# BTZen - library to asynchronously access Bluetooth devices.
#
# Copyright (C) 2015-2022 by Artur Wroblewski <wrobell@riseup.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""
Basic enums, records and types.
"""

import dataclasses as dtc
import enum
import typing as tp

T = tp.TypeVar('T')
Converter = tp.Callable[[bytes], T]
AnyTrigger = tp.Union['Trigger', 'NoTrigger']

class Make(enum.Enum):
    """
    Bluetooth device make.
    """
    STANDARD = 'standard'
    SENSOR_TAG = 'sensor_tag'
    THINGY52 = 'thingy52'
    OSTC = 'ostc'
    MI_SMART_SCALE = 'mi_smart_scale'

class ServiceType(enum.Enum):
    """
    Bluetooth service type.
    """
    ACCELEROMETER = enum.auto()
    BUTTON = enum.auto()
    BATTERY_LEVEL = enum.auto()
    HUMIDITY = enum.auto()
    LIGHT = enum.auto()
    LIGHT_RGB = enum.auto()
    PRESSURE = enum.auto()
    SERIAL = enum.auto()
    TEMPERATURE = enum.auto()
    WEIGHT_MEASUREMENT = enum.auto()

class AddressType(enum.Enum):
    """
    Bluetooth device address type.


    .. seealso::

        `ConnectDevice` method documentation at
        https://git.kernel.org/pub/scm/bluetooth/bluez.git/tree/doc/adapter-api.txt.
    """
    PUBLIC = 'public'
    RANDOM = 'random'

class TriggerCondition(enum.IntEnum):
    """
    Condition value for Bluetooth Environmental Sensing device trigger
    information.

    Use `NoTrigger` for inactive trigger.

    NOTE: Incomplete, see Bluetooth Environmental Sensing Service
        specification, page 18.
    """
    FIXED_TIME = 0x01
    ON_CHANGE = 0x04

@dtc.dataclass(frozen=True)
class Trigger:
    """
    Bluetooth Environmental Sensing device trigger information.
    """
    condition: TriggerCondition
    operand: tp.Optional[tp.Any]=None

@dtc.dataclass
class NoTrigger:
    """
    Singleton object indicating no trigger for a Bluetooth device.
    """

class Button(enum.IntFlag):
    """
    Base enumeration class for state of buttons found on various Bluetooth
    devices.
    """

@dtc.dataclass(frozen=True)
class LightColor:
    """
    Light value with RGB colour information.
    """
    red: float
    blue: float
    green: float
    clear: float

class WeightFlags(enum.IntFlag):
    IMPERIAL = 0x01
    TIMESTAMP = 0x02
    USER_ID = 0x04
    BMI = 0x08
    RESERVED_1 = 0x10
    RESERVED_2 = 0x20
    RESERVED_3 = 0x40
    RESERVED_4 = 0x80

@dtc.dataclass(frozen=True)
class WeightData:
    """
    Weight measurement data.

    var flags: Weight scale flags value.
    var weight: Weight measurement value.
    """
    flags: WeightFlags
    weight: float

# vim: sw=4:et:ai



#
# BTZen - library to asynchronously access Bluetooth devices.
#
# Copyright (C) 2015-2022 by Artur Wroblewski <wrobell@riseup.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from importlib.metadata import version

# register devices
from . import serial as mod_serial
from . import sensortag
from . import thingy52
from . import bluez

from .btweight import MiScaleWeightData
from .data import AddressType, Button, Make, LightColor, TriggerCondition, \
    Trigger, ServiceType, WeightFlags, WeightData
from .device import DeviceBase, Device, DeviceTrigger, set_interval, \
    set_trigger, set_address_type, create_device, \
    accelerometer, battery_level, button, humidity, light, \
    light_rgb, pressure, serial, temperature, weight
from .devio import read, read_all, write, enable, disable
from .serial import SerialService
from .service import Service, ServiceCharacteristic, ServiceInterface
from .cm import connect
from .error import *
from .session import is_active
from .sensortag import SensorTagButtonState
from .thingy52 import Thingy52ButtonState

__version__ = version('btzen')

__all__ = [
    # connection session
    'is_active', 'connect',

    # bluetooth device descriptors and constructors
    'DeviceBase', 'Device', 'DeviceTrigger', 'create_device',
    'accelerometer', 'battery_level', 'button', 'humidity', 'light',
    'light_rgb', 'pressure', 'serial', 'temperature', 'weight',

    # bluetooth service descriptors
    'Service', 'ServiceCharacteristic', 'ServiceInterface', 'SerialService',

    # basic data
    'AddressType', 'Make', 'ServiceType', 'Button', 'LightColor', 'WeightData',
    'WeightFlags', 'Trigger', 'TriggerCondition',

    # bluetooth device i/o
    'read', 'read_all', 'write',

    # bluetooth device configuration
    'set_interval', 'set_trigger', 'set_address_type',

    # make specific objects
    'SensorTagButtonState', 'Thingy52ButtonState', 'MiScaleWeightData',
]

# vim: sw=4:et:ai



#
# BTZen - library to asynchronously access Bluetooth devices.
#
# Copyright (C) 2015-2022 by Artur Wroblewski <wrobell@riseup.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""
Bluetooth services implemented by BlueZ protocol stack.
"""

from .data import Make, ServiceType, Trigger, TriggerCondition
from .service import ServiceInterface, register_service
from .util import to_uuid

register_service(
    Make.STANDARD,
    ServiceType.BATTERY_LEVEL,
    ServiceInterface(
        to_uuid(0x180f),
        'org.bluez.Battery1',
        'Percentage',
        'y'
    ),
    trigger=Trigger(TriggerCondition.ON_CHANGE),
)

# vim: sw=4:et:ai



#
# BTZen - library to asynchronously access Bluetooth devices.
#
# Copyright (C) 2015-2022 by Artur Wroblewski <wrobell@riseup.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""
Connection manager to manage connections of multiple Bluetooth devices.

Based on

    https://gist.github.com/parthitce/eb6b751df3235f7247babc4c9aba41d8

NOTE: Discovery is still required on devices like Raspberry Pi to reconnect
    a long running device.

When starting and running connection manager

1. Register Bluetooth agent.
2. Create connection manager object on D-Bus event bus and register UUIDs
   of services of devices to be managed by the connection manager.
3. For each device
3.1. Remove device preemptively to allow new connection.
3.2. Use `ConnectDevice` method of adapter interface to connect to the
     device. If fails, then move to 3.1 above.
3.3. Set the device to be seen as trusted.
3.4. Wait for `ServicesResolved` property to be changed.
3.4.1. If the property is set to true enable Bluetooth device.
3.4.2. If the property is set to false, then disable Bluetooth device.

When connection manager is closed

1. Close each Bluetooth device.
2. Disconnect each Bluetooth device.
3. Remove each Bluetooth device.
4. Unregister Bluetooth agent.
5. Close connection manager.

"""

import asyncio
import logging
import typing as tp
from collections import defaultdict
from collections.abc import Coroutine, AsyncGenerator, AsyncIterator
from contextlib import asynccontextmanager
from contextvars import ContextVar
from functools import partial
from operator import attrgetter

from . import _cm  # type: ignore
from .bus import Bus
from .error import BTZenError
from .config import DEFAULT_DBUS_TIMEOUT, DEFAULT_CONNECTION_TIMEOUT
from .data import T, AddressType
from .device import DeviceBase
from .devio import enable, disable, disarm
from .session import BT_SESSION, Session, get_session, is_active
from .service import S
from .util import concat

Devices = tp.Iterable[DeviceBase[S, T]]

logger = logging.getLogger(__name__)


@asynccontextmanager
async def connect(
        devices: Devices,  # type: ignore
        *,
        interface: str='hci0'
    ) -> AsyncIterator[Session]:

    bus = Bus.create_bus(interface)
    adapter_path = bus.adapter_path()

    session = Session(bus)
    BT_SESSION.set(session)

    # TODO: use context variable
    _dbus_timeout = DEFAULT_DBUS_TIMEOUT
    await _cm.bt_register_agent(bus.system_bus, _dbus_timeout)

    by_mac = defaultdict(set)
    for dev in devices:
        by_mac[dev.mac].add(dev)

    # TODO: if bluez daemon is restarted, the connection manager needs
    # to be reinitialized
    bt_services = set(dev.service.uuid for dev in concat(by_mac.values()))
    cm_handle = await _cm.cm_init(bus.system_bus, adapter_path, bt_services)

    for mac, devs in by_mac.items():
        task = manage_connection(bus, mac, devs)
        session.add_connection_task(mac, task)

    session.start()
    try:
        yield session
    finally:
        session.stop()

        await disarm(
            'agent unregistered',
            'agent failed to unregister',
            _cm.bt_unregister_agent,
            bus.system_bus
        )

        await disarm(
            'connection manager unregistered',
            'connection manager failed to unregister',
            _cm.cm_close,
            bus.system_bus,
            adapter_path,
            cm_handle
        )

async def manage_connection(bus: Bus, mac: str, devices: Devices) -> None:  # type: ignore
    """
    Manage Bluetooth connection for the devices.
    """
    # determine connection address type and favour random one
    address_types = set(dev.address_type for dev in devices)
    has_random = AddressType.RANDOM in address_types
    address_type = AddressType.RANDOM if has_random else AddressType.PUBLIC

    # enable monitoring of the `ServicesResolved` property first
    bus._dev_property_start(mac, 'ServicesResolved')

    # create connection to a device; if it exists already, then remove it
    # first; if it exists in bluez daemon registry, then creating new
    # connection blocks
    created = False
    while not created and is_active():
        await remove_connection(bus, mac)
        created = await create_connection(bus, mac, address_type)

    try:
        await restart_devices(bus, mac, devices)
    finally:
        bus._dev_property_stop(mac, 'ServicesResolved')

        await disarm(
            'device {} disconnected'.format(mac),
            'device {} failed to disconnect'.format(mac),
            _cm.bt_disconnect,
            bus.system_bus,
            bus.dev_path(mac),
        )

        await remove_connection(bus, mac)

# TODO: make real async
async def remove_connection(bus: Bus, mac: str) -> None:
    await disarm(
        'connection for device {} removed'.format(mac),
        'removal of connection failed for device {}'.format(mac),
        _cm.bt_remove,
        bus.system_bus,
        bus.adapter_path(),
        bus.dev_path(mac)
    )

async def create_connection(
        bus: Bus,
        mac: str,
        address_type: AddressType,
    ) -> bool:
    # create connection to a device
    #
    # NOTE: bluez 5.50 - scanning by external programs shall be off or we
    # will never connect

    _dbus_timeout = DEFAULT_DBUS_TIMEOUT

    created = False
    dev_path = bus.dev_path(mac)
    try:
        logger.info(
            'connect device {} via controller {}, address type {}'
            .format(mac, bus.adapter_path(), address_type)
        )
        await _cm.bt_connect(
            bus.system_bus, bus.adapter_path(), mac, address_type.value,
            DEFAULT_CONNECTION_TIMEOUT
        )
    except (BTZenError, asyncio.CancelledError) as ex:
        is_active = get_session().is_active()
        if is_active and str(ex) == 'Already Exists':
            created = True
        elif is_active:
            sleep = int(_dbus_timeout / 1e6)
            logger.info(
                'connection for {} failed: {}, sleep for {}s'
                .format(mac, ex, sleep)
            )
            await asyncio.sleep(sleep)
    else:
        created = True

    if created:
        _cm.bt_device_set_trusted(bus.system_bus, dev_path)
    return created

async def enable_devices(mac: str, devices: Devices) -> None:  # type: ignore
    logger.info('enabling devices: {}'.format(mac))

    for dev in devices:
        await enable(dev)

    get_session().set_connected(mac)
    logger.info('enabled services: {}'.format(mac))

async def disable_devices(mac: str, devices: Devices) -> None:  # type: ignore
    logger.info('disabling services: {}'.format(mac))

    session = get_session()

    # clear connection flag as soon as possible, to prevent reading from
    # disabled device
    session.set_disconnected(mac)
    session.cancel_device_tasks(mac, 'Disable device')

    for dev in devices:
        # no exception checks as the disable functions should not raise
        # exceptions on failure
        await disable(dev)

    logger.info('disabled services: {}'.format(mac))

async def restart_devices(bus: Bus, mac: str, devices: Devices) -> None:  # type: ignore
    """
    Enable or disable Bluetooth device when property 'ServicesResolved`
    changes.
    """
    # the `ServicesResolved` property monitoring is started by a caller, so
    # just wait for services to be resolved
    async for resolved in resolve_services(bus, mac, devices):
        try:
            enabled = False
            if resolved:
                await enable_devices(mac, devices)
                enabled = True
        except asyncio.CancelledError as ex:
            logger.info(
                'enabling devices for %s failed, seems to be not connected',
                mac
            )
            if not is_active():
                raise
        finally:
            # not resolved or not all devices enabled, then disable all;
            # some devices might be partially disabled and they still need
            # release other resources, i.e. d-bus related
            if not enabled:
                # some devices might be enabled, so try to disable them
                await disable_devices(mac, devices)

async def resolve_services(
        bus: Bus, mac: str, devices: Devices  # type: ignore
    ) -> AsyncGenerator[bool, None]:
    """
    Asynchronous generator waiting for a Bluetooth device to be
    resolved.
    """
    while True:
        logger.info(
            'device {} waiting for services resolved status change'
            .format(mac)
        )
        resolved = await bus._dev_property_get(mac, 'ServicesResolved')
        logger.info('device {} services resolved: {}'.format(mac, resolved))
        yield resolved

# vim: sw=4:et:ai



#
# BTZen - library to asynchronously access Bluetooth devices.
#
# Copyright (C) 2015-2022 by Artur Wroblewski <wrobell@riseup.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""
BTZen utility functions.
"""

import typing as tp
from functools import partial
from itertools import chain

concat = chain.from_iterable
to_int = partial(int.from_bytes, byteorder='little')

# function to convert 16-bit UUID to full 128-bit Bluetooth normative UUID
# string
to_uuid: tp.Callable[[int], str] = '0000{:04x}-0000-1000-8000-00805f9b34fb'.format

# vim: sw=4:et:ai



#
# BTZen - library to asynchronously access Bluetooth devices.
#
# Copyright (C) 2015-2022 by Artur Wroblewski <wrobell@riseup.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""
Driver for serial device over Bluetooth Smart connection. Implements
Stollmann protcol

    http://www.telit.com/fileadmin/user_upload/products/Downloads/sr-rf/BlueMod/TIO_Implementation_Guide_r04.pdf

Tested with HeinrichsWeikamp OSTC 2 dive computer.
"""

import asyncio
import math
import logging
import typing as tp
from binascii import hexlify
from contextlib import asynccontextmanager
from functools import partial, cache

from . import _btzen  # type: ignore
from .bus import Bus
from .config import DEFAULT_DBUS_TIMEOUT
from .data import Make, ServiceType
from .device import Device
from .devio import enable, disable, read, write, disarm
from .service import Service, register_service
from .session import get_session, connected
from .util import to_uuid

logger = logging.getLogger(__name__)

T = tp.TypeVar('T', bound=bytes)

UUID_RX_UART = '00000001-0000-1000-8000-008025000000'
UUID_TX_UART = '00000002-0000-1000-8000-008025000000'
UUID_TX_CREDIT = '00000004-0000-1000-8000-008025000000'
UUID_RX_CREDIT = '00000003-0000-1000-8000-008025000000'

class State(tp.TypedDict):
    """
    Device state with buffer and number of RX credits.
    """
    buffer: bytearray
    rx_credits: int

class SerialService(Service):
    pass

register_service(
    Make.OSTC,
    ServiceType.SERIAL,
    SerialService(to_uuid(0xfefb)),
)

@cache
def device_state(device: Device[SerialService, T]) -> State:
    """
    Create or get serial device state.
    """
    return State(buffer=bytearray(), rx_credits=0)

@read.register  # type: ignore
async def _read_serial(device: Device[SerialService, T], n: int) -> bytes:
    async with connected(device) as session:
        task = session.create_future(device, _read_data(session.bus, device, n))   # type: ignore
        return (await task)  # type: ignore

async def _read_data(bus: Bus, device: Device[SerialService, T], n: int) -> bytes:
    state = device_state(device)
    data = bytearray(state['buffer'])

    await bus.ensure_characteristic_paths(
        device.mac, UUID_TX_UART, UUID_RX_UART, UUID_TX_CREDIT, UUID_RX_CREDIT
    )

    path_uart = bus.characteristic_path(device.mac, UUID_TX_UART)

    while len(data) < n:
        async with _rx_credits_mgr(bus, device, n - len(data)):
            item = await bus._gatt_get(path_uart)
            data.extend(item)

            if __debug__:
                logger.debug(
                    'bytes read {}, last {}, tx credits size {}'
                    .format(len(data), hexlify(data[-5:]).decode(), _tx_credit_size(bus, device.mac))
                )

    assert len(data) >= n

    # keep extra data in buffer, return only requested number of bytes
    state['buffer'] = data[n:]
    return data[:n]

@write.register  # type: ignore
async def _write_serial(device: Device[SerialService, T], data: bytes) -> None:
    assert len(data) <= 20
    async with connected(device) as session:
        state = device_state(device)

        if state['rx_credits'] < 1:
            await _add_rx_credits(session.bus, device)

        if _tx_credit_size(session.bus, device.mac) > 0:
            logger.debug('requesting tx credits on write')
            await _tx_credit(session.bus, device.mac)

        await _write(session.bus, device.mac, UUID_RX_UART, data)

@enable.register  # type: ignore
async def _enable_serial(device: Device[SerialService, T]) -> None:
    bus = get_session().bus
    get_path = partial(bus.characteristic_path, device.mac)

    # reset cache for device
    state = device_state(device)
    state['buffer'] = bytearray()
    state['rx_credits'] = 0

    bus._gatt_start(get_path(UUID_TX_CREDIT))
    bus._gatt_start(get_path(UUID_TX_UART))

    logger.debug('requesting xx credits on enable')
    await _add_rx_credits(bus, device)

    logger.debug('requesting tx credits on enable')
    await _tx_credit(bus, device.mac)

@disable.register  # type: ignore
async def _disable_serial(device: Device[SerialService, T]) -> None:
    bus = get_session().bus
    await disarm(
        'disabled tx credit for {}'.format(device.mac),
        'cannot disabl tx credit for {}'.format(device.mac),
        bus._gatt_stop,
        bus.characteristic_path(device.mac, UUID_TX_CREDIT),
    )
    await disarm(
        'disabled tx uart for {}'.format(device.mac),
        'cannot disabl tx uart for {}'.format(device.mac),
        bus._gatt_stop,
        bus.characteristic_path(device.mac, UUID_TX_UART),
    )

async def _tx_credit(bus: Bus, mac: str) -> None:
    path = bus.characteristic_path(mac, UUID_TX_CREDIT)
    n = await bus._gatt_get(path)
    logger.debug('got tx credits: {}'.format(n))

def _tx_credit_size(bus: Bus, mac: str) -> int:
    path = bus.characteristic_path(mac, UUID_TX_CREDIT)
    return bus._gatt_size(path)

async def _add_rx_credits(
        bus: Bus, device: Device[SerialService, T], n: int=0x20
    ) -> None:

    state = device_state(device)
    await _write(bus, device.mac, UUID_RX_CREDIT, bytes([n]))
    state['rx_credits'] += n
    logger.debug('rx credits: {}'.format(state['rx_credits']))

@asynccontextmanager
async def _rx_credits_mgr(
        bus: Bus, device: Device[SerialService, T], n: int
    ) -> tp.AsyncIterator[None]:

    state = device_state(device)
    if state['rx_credits'] < 1:
        await _add_rx_credits(bus, device, credits_for(n))

    try:
        yield
    finally:
        state['rx_credits'] -= 1

async def _write(bus: Bus, mac: str, uuid: str, data: bytes) -> None:
    path = bus.characteristic_path(mac, uuid)
    task = _btzen.bt_write(bus.system_bus, path, data, DEFAULT_DBUS_TIMEOUT)
    await task

def credits_for(n: int) -> int:
    """
    Calculate number of required credits to send `n` number of bytes.
    """
    return min(255, math.ceil(n / 20))

# vim: sw=4:et:ai



#
# BTZen - library to asynchronously access Bluetooth devices.
#
# Copyright (C) 2015-2022 by Artur Wroblewski <wrobell@riseup.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""
Nordic Thingy:52 Bluetooth device sensors

    https://nordicsemiconductor.github.io/Nordic-Thingy52-FW/documentation/firmware_architecture.html

The sensors do not implement the Bluetooth Environmental Sensing interfaces
and require custom classes. All Thingy:52 Bluetooth device sensors are
notifying.
"""

import dataclasses as dtc
import logging
import struct
import typing as tp
from collections import defaultdict
from functools import partial, cache

from .data import T, AddressType, Button, LightColor, Make, ServiceType, \
    Trigger, TriggerCondition
from .device import DeviceTrigger, set_trigger
from .devio import enable, _enable_dev_trigger, write_config
from .service import register_service, ServiceCharacteristic
from .util import to_int

logger = logging.getLogger(__name__)

CONFIG_DATA_FMT = struct.Struct('<HHHHBBBB')
SENSOR_COLOR_FMT = struct.Struct('<HHHH')
LIGHT_MAX = 0xffff

# function to convert 16-bit UUID to full 128-bit Thingy:52 UUID
to_uuid = 'ef68{:04x}-9b35-4933-9b10-52ffa9740042'.format

register_th = partial(
    register_service,
    Make.THINGY52,
    address_type=AddressType.RANDOM,
    trigger=Trigger(TriggerCondition.FIXED_TIME, 1),
)

@dtc.dataclass(frozen=True)
class Thingy52Service(ServiceCharacteristic):
    uuid_conf: str
    config_entry: str

@dtc.dataclass
class Thingy52Config:
    """
    Thingy:52 Bluetooth device configuration for the weather service
    sensors.
    """
    # temperature sensor data read interval
    temperature: float = 1.0
    # pressure sensor data read interval
    pressure: float = 1.0
    # humidity sensor data read interval
    humidity: float = 1.0
    # color sensor data read interval
    color: float = 1.0
    # gas sensor data read interval (1 - 1s, 2 - 10s, 3 - 60s)
    gas: int = 1
    # color sensor LED calibration (RGB value)
    rgb: tuple[int, int, int] = (0, 255, 0)

_CONFIG_CACHE: defaultdict[str, Thingy52Config] = defaultdict(Thingy52Config)

class Thingy52ButtonState(Button):
    """
    Thingy:52 Bluetooth device button state.
    """
    OFF = 0x00
    ON = 0x01

def convert_light(data: bytes) -> LightColor:
    """
    Convert data of BH1745 light sensor in Thingy:52 to light color object.
    """
    values = SENSOR_COLOR_FMT.unpack(data)
    return LightColor(*(v / LIGHT_MAX for v in values))

register_th(
    ServiceType.PRESSURE,
    Thingy52Service(
        to_uuid(0x0200),
        to_uuid(0x0202),
        5,
        to_uuid(0x0206),
        'pressure',
    ),
    convert=lambda data: to_int(data[:4]) * 100 + data[4],
)

register_th(
    ServiceType.TEMPERATURE,
    Thingy52Service(
        to_uuid(0x0200),
        to_uuid(0x0201),
        2,
        to_uuid(0x0206),
        'temperature',
    ),
    convert=lambda data: int.from_bytes(data[0:1], signed=True) + data[1] / 100
)

register_th(
    ServiceType.HUMIDITY,
    Thingy52Service(
        to_uuid(0x0200),
        to_uuid(0x0203),
        1,
        to_uuid(0x0206),
        'humidity',
    ),
    convert=lambda data: data[0],
)

register_th(
    ServiceType.LIGHT_RGB,
    Thingy52Service(
        to_uuid(0x0200),
        to_uuid(0x0205),
        8,
        to_uuid(0x0206),
        'color',
    ),
    convert=convert_light,
)

register_th(
    ServiceType.BUTTON,
    ServiceCharacteristic(
        to_uuid(0x0300),
        to_uuid(0x0302),
        1,
    ),
    convert=lambda data: Thingy52ButtonState(data[0]),
)

@enable.register  # type: ignore
async def _enable_thingy52(device: DeviceTrigger[Thingy52Service, T]) -> None:
    to_ms: tp.Callable[[float], int] = lambda v: int(v * 1000)
    mac = device.mac
    srv = device.service

    config = _CONFIG_CACHE[mac]
    data = CONFIG_DATA_FMT.pack(
        to_ms(config.temperature),
        to_ms(config.pressure),
        to_ms(config.humidity),
        to_ms(config.color),
        config.gas,
        *config.rgb,
    )
    await write_config(mac, srv.uuid_conf, data)
    await _enable_dev_trigger(device)

@set_trigger.register  # type: ignore
def _set_trigger_thingy52(
        device: DeviceTrigger[Thingy52Service, T],
        condition: TriggerCondition,
        *,
        operand: tp.Optional[float]=None,
    ) -> DeviceTrigger[Thingy52Service, T]:

    assert operand is not None

    mac = device.mac
    config_entry = device.service.config_entry
    config = _CONFIG_CACHE[device.mac]
    config = dtc.replace(config, **{config_entry: operand})
    _CONFIG_CACHE[mac] = config

    return device

# class DeviceThingy52Configuration(DeviceCharacteristic):
#     """
#     Thingy:52 Bluetooth device configuration.
#     """
#     ADDRESS_TYPE = 'random'
# 
# class ConnectionParameters(DeviceThingy52Configuration):
#     info = InfoCharacteristic(
#         to_uuid(0x0100),
#         to_uuid(0x0104),
#         8,
#     )
#     async def set_params(
#         self,
#         min_conn_interval,
#         max_conn_interval,
#         slave_latency,
#         supervision_timeout,
#     ):
#         system_bus = self._bus.system_bus
#         data = struct.pack(
#             '<HHHH',
#             min_conn_interval,
#             max_conn_interval,
#             slave_latency,
#             supervision_timeout,
#         )
#         path = self._get_path(self.info.uuid_data)
#         await _btzen.bt_write(system_bus, path, data)

# vim: sw=4:et:ai



#
# BTZen - library to asynchronously access Bluetooth devices.
#
# Copyright (C) 2015-2022 by Artur Wroblewski <wrobell@riseup.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from __future__ import annotations

import asyncio
import contextvars
import logging
import typing as tp
from functools import partial

from . import _btzen  # type: ignore
from . import _sd_bus  # type: ignore
from .config import DEFAULT_CHARACTERISTIC_PATH_RETRY
from .error import BTZenError, ConnectionError

logger = logging.getLogger(__name__)

INTERFACE_DEVICE = 'org.bluez.Device1'
INTERFACE_GATT_CHR = 'org.bluez.GattCharacteristic1'

def _mac_to_path(mac: str) -> str:
    return mac.replace(':', '_').upper()

class Bus:
    BUS = contextvars.ContextVar[tp.Optional['Bus']]('BUS', default=None)

    def __init__(self, system_bus: _sd_bus.Bus, interface: str) -> None:
        self.system_bus = system_bus
        self.interface = interface

        loop = asyncio.get_event_loop()
        process = partial(_btzen.bt_process, system_bus)
        loop.add_reader(system_bus.fileno, process)

        self._notifications = Notifications(self)
        self._characteristic_cache: dict[tuple[str, str], str] = {}

    @staticmethod
    def create_bus(interface: str) -> Bus:
        """
        Create system bus reference.

        The reference is local to current context.

        :param interface: Bluetooth interface, i.e. `hci0`.
        """
        bus = Bus.BUS.get()
        if bus is None:
            system_bus = _sd_bus.default_bus()
            bus = Bus(system_bus, interface)
            Bus.BUS.set(bus)
        else:
            raise ValueError('Bus already exists')

        return bus

    @staticmethod
    def get_bus() -> Bus:
        """
        Get system bus reference from current context.
        """
        bus = Bus.BUS.get()
        if bus is None:
            raise ValueError('Bus does not exist')
        return bus

    def characteristic_path(self, mac: str, uuid: str) -> str:
        """
        Get Bluetooth GATT Characteristic path for Bluetooth device address
        and GATT Characteristic UUID.

        The paths are cached.

        :param mac: Bluetooth device MAC address.
        :param uuid: UUID of Bluetooth GATT Characteristic.
        """
        key = mac, uuid
        prefix = self.dev_path(mac)
        path = self._characteristic_cache.get(key)
        if path is None:
            path = _btzen.bt_characteristic(self.system_bus, prefix, uuid)

        # store in cache only non-null values
        if path is None:
            raise BTZenError('Path for {}/{} not found'.format(mac, uuid))
        else:
            self._characteristic_cache[key] = path
        return path

    async def ensure_characteristic_paths(self, mac: str, *uuid: str) -> None:
        """
        Ensure Bluetooth GATT characteristic path exists for each UUID.
        """
        # TODO: in the future we might want to use object manager
        for u in uuid:
            await self.ensure_characteristic_path(mac, u)

    async def ensure_characteristic_path(self, mac: str, uuid: str) -> None:
        key = mac, uuid
        prefix = self.dev_path(mac)
        for i in range(DEFAULT_CHARACTERISTIC_PATH_RETRY):
            path = _btzen.bt_characteristic(self.system_bus, prefix, uuid)
            if path is None:
                logger.warning(
                    'characteristic path not found for {}/{}'.format(mac, uuid)
                )
                await asyncio.sleep(1)
            else:
                self._characteristic_cache[key] = path
                return

        raise BTZenError(
            'Cannot determine characteristic path for {}/{}'.format(mac, uuid)
        )

    def _gatt_start(self, path: str) -> None:
        # TODO: creates notification session; if another session started,
        # then we get notifications twice; this needs to be fixed
        self._notifications.start(path, INTERFACE_GATT_CHR, 'Value')
        try:
            _btzen.bt_notify_start(self.system_bus, path)
        except Exception as ex:
            logger.warning('cannot start notification for {}: {}'.format(path, ex))
            self._notifications.stop(path, INTERFACE_GATT_CHR)
            raise

    def adapter_path(self) -> str:
        return '/org/bluez/{}'.format(self.interface)

    def dev_path(self, mac: str) -> str:
        return '/org/bluez/{}/dev_{}'.format(self.interface, _mac_to_path(mac))

    async def _gatt_get(self, path: str) -> tp.Any:
        task = self._notifications.get(path, INTERFACE_GATT_CHR, 'Value')
        return (await task)

    def _gatt_stop(self, path: str) -> None:
        try:
            _btzen.bt_notify_stop(self.system_bus, path)
        except Exception as ex:
            logger.warning('cannot stop notification for {}: {}'.format(path, ex))
        finally:
            self._notifications.stop(path, INTERFACE_GATT_CHR)

    def _gatt_size(self, path: str) -> int:
        return self._notifications.size(path, INTERFACE_GATT_CHR, 'Value')

    def _dev_property_start(
        self, mac: str, name: str, iface: str=INTERFACE_DEVICE
    ) -> None:
        path = self.dev_path(mac)
        self._notifications.start(path, iface, name)

    async def _dev_property_get(
            self,
            mac: str,
            name: str,
            iface: str=INTERFACE_DEVICE,
    ) -> tp.Any:
        path = self.dev_path(mac)
        value = await self._notifications.get(path, iface, name)
        return value

    def _dev_property_stop(
        self, mac: str, name: str, iface: str=INTERFACE_DEVICE
    ) -> None:
        path = self.dev_path(mac)
        self._notifications.stop(path, iface)

    async def _property(
        self, mac: str, iface: str, name: str, type: str='s'
    ) -> tp.Any:
        bus = self.system_bus
        path = self.dev_path(mac)
        value = await _btzen.bt_property(bus, path, iface, name, type)
        return value

    async def _get_name(self, mac: str) -> str:
        path = self.dev_path(mac)
        bus = self.system_bus
        value = await _btzen.bt_property(bus, path, INTERFACE_DEVICE, 'Name', 's')
        return value  # type: ignore

class Notifications:
    def __init__(self, bus: Bus):
        self._data: dict[tuple[str, str], _btzen.PropertyNotification] = {}
        self._bus = bus

    def start(self, path: str, iface: str, name: str) -> None:
        key = path, iface
        bus = self._bus.system_bus

        assert key not in self._data
        data = _btzen.bt_property_monitor_start(bus, path, iface)
        self._data[key] = data

        assert key in self._data
        data.register(name)

    def size(self, path: str, iface: str, name: str) -> int:
        key = path, iface
        return self._data[key].size(name)  # type: ignore

    async def get(self, path: str, iface: str, name: str) -> tp.Any:
        key = path, iface
        data = self._data[key]
        return (await data.get(name))

    def stop(self, path: str, iface: str) -> None:
        # TODO: add name and call PropertyNotification.stop when no
        # properties monitored
        key = path, iface
        if key in self._data:
            self._data[key].stop()
            del self._data[key]

        assert key not in self._data

# vim: sw=4:et:ai



#
# BTZen - library to asynchronously access Bluetooth devices.
#
# Copyright (C) 2015-2022 by Artur Wroblewski <wrobell@riseup.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""
Texas Instrument Sensor Tag Bluetooth device sensors.

The sensors do not implement the Bluetooth Environmental Sensing interfaces
and require some custom implementation.

The sensors on device read data at certain time interval (period), usually
one second. Functions `set_interval` or `set_trigger` change this interval.
To change the time interval for a device with inactive trigger, use
`dataclasses.replace` function.

The identificators for specific sensors can be found at

CC2541DK
    http://processors.wiki.ti.com/index.php/SensorTag_User_Guide
CC2650STK
    http://processors.wiki.ti.com/index.php/CC2650_SensorTag_User's_Guide
"""

import asyncio
import dataclasses as dtc
import enum
import logging
import struct
import typing as tp
from functools import partial

from .data import T, Button, Make, ServiceType, Trigger, TriggerCondition
from .device import Device, DeviceTrigger
from .devio import enable, disable, write_config, disarm, \
    _enable_dev_trigger
from .service import ServiceCharacteristic, register_service
from .session import get_session
from .util import to_int, to_uuid as to_bt_uuid

logger = logging.getLogger(__name__)

HDC1000_HUMIDITY = 65536 / 100
ACCEL = 0x08 | 0x10 | 0x20
ACCEL_WAKE_ON_MOTION = 0x80
MPU9250_GYRO = 65536 / 500
MPU9250_ACCEL_2G = 32768 / 2
MPU9250_ACCEL_UNPACK = struct.Struct('<3h').unpack

@dtc.dataclass(frozen=True)
class SensorTagService(ServiceCharacteristic):
    """
    Bluetooth service descriptor for GATT characteristics of Texas
    Instrument Sensor Tag Bluetooth device sensors.

    :var uuid_conf: UUID of characteristic to write and read device
        configuration.
    :var uuid_trigger: UUID of characteristic to write and read device
        trigger data.
    :var config_on: Default configuration of device to switch device on.
    :var config_off: Default configuration of device to switch device off.
    :var interval: Time interval (period) of sensor.
    """
    uuid_conf: str
    uuid_trigger: str
    config_on: bytes
    config_off: bytes
    interval: float=1

class SensorTagButtonState(Button):
    """
    Sensor Tag Bluetooth device button state.
    """
    OFF = 0x00
    USER = 0x01
    POWER = 0x02
    REED_RELAY = 0x04

# function to convert 16-bit UUID to full 128-bit Sensor Tag UUID
to_uuid = 'f000{:04x}-0451-4000-b000-000000000000'.format

register_st = partial(register_service, Make.SENSOR_TAG)

def convert_light(data: bytes) -> float:
    """
    Convert Sensor Tag light sensor data to a lux value.
    """
    v = to_int(data)
    m = (v & 0x0FFF) / 100
    e = (v & 0xF000) >> 12
    return m * (2 << e)

def convert_accel(data: bytes) -> tuple[float, float, float]:
    """
    Convert Sensor Tag Bluetooth device accelerometer data into (x, y, z)
    values.
    """
    # gyro: data[:6]
    # magnet: data[12:]
    x, y, z = tp.cast(tuple[float, float, float], MPU9250_ACCEL_UNPACK(data[6:12]))
    return (x / MPU9250_ACCEL_2G, y / MPU9250_ACCEL_2G, z / MPU9250_ACCEL_2G)

def convert_button(data: bytes) -> SensorTagButtonState:
    """
    Convert Sensor Tag Bluetooth device button data into button state
    object.
    """
    return SensorTagButtonState(data[0])

register_st(
    ServiceType.PRESSURE,
    SensorTagService(
        to_uuid(0xaa40),
        to_uuid(0xaa41),
        6,
        to_uuid(0xaa42),
        to_uuid(0xaa44),
        b'\x01',
        b'\x00',
    ),
    convert=lambda v: to_int(v[3:]),
)

register_st(
    ServiceType.TEMPERATURE,
    SensorTagService(
        to_uuid(0xaa00),
        to_uuid(0xaa01),
        4,
        to_uuid(0xaa02),
        to_uuid(0xaa03),
        b'\x01',
        b'\x00',
    ),
    convert=lambda v: to_int(v[2:]) / 128,
)

register_st(
    ServiceType.HUMIDITY,
    SensorTagService(
        to_uuid(0xaa20),
        to_uuid(0xaa21),
        4,
        to_uuid(0xaa22),
        to_uuid(0xaa23),
        b'\x01',
        b'\x00',
    ),
    convert=lambda v: to_int(v[2:]) / HDC1000_HUMIDITY,
)

register_st(
    ServiceType.LIGHT,
    SensorTagService(
        to_uuid(0xaa70),
        to_uuid(0xaa71),
        2,
        to_uuid(0xaa72),
        to_uuid(0xaa73),
        b'\x01',
        b'\x00',
    ),
    convert=convert_light,
)

register_st(
    ServiceType.ACCELEROMETER,
    SensorTagService(
        to_uuid(0xaa80),
        to_uuid(0xaa81),
        18,
        to_uuid(0xaa82),
        to_uuid(0xaa83),
        struct.pack('<H', ACCEL | ACCEL_WAKE_ON_MOTION),
        b'\x00\x00',
    ),
    convert=convert_accel,
    trigger=Trigger(TriggerCondition.FIXED_TIME, 0.1),
)

register_st(
    ServiceType.BUTTON,
    ServiceCharacteristic(
        to_bt_uuid(0xffe0),
        to_bt_uuid(0xffe1),
        1,
    ),
    convert=convert_button,
    trigger=Trigger(TriggerCondition.ON_CHANGE),
)

@enable.register  # type: ignore
async def _enable_sensor_tag(device: Device[SensorTagService, T]) -> None:
    bus = get_session().bus
    srv = device.service

    await bus.ensure_characteristic_paths(
        device.mac, srv.uuid_data, srv.uuid_conf, srv.uuid_trigger
    )
    await write_config(device.mac, srv.uuid_conf, srv.config_on)

    value = int(device.service.interval * 100)
    assert value < 256  # TODO: raise value error

    await write_config(device.mac, device.service.uuid_trigger, bytes([value]))
    logger.info('interval for {} is set'.format(device))

@disable.register  # type: ignore
async def _disable_sensor_tag(device: Device[SensorTagService, T]) -> None:
    srv = device.service
    await disarm(
        '{} disabled'.format(device),
        'cannot disable {}'.format(device),
        write_config, device.mac, srv.uuid_conf, srv.config_off
    )

@enable.register  # type: ignore
async def _enable_sensor_tag_tr(
        device: DeviceTrigger[SensorTagService, T]
    ) -> None:

    await _enable_sensor_tag(device)

    assert device.trigger.operand is not None
    value = int(device.trigger.operand * 100)
    assert value < 256  # TODO: raise value error

    await write_config(device.mac, device.service.uuid_trigger, bytes([value]))
    logger.info('trigger for {} is set'.format(device))

    await _enable_dev_trigger(device)

@disable.register  # type: ignore
async def _disable_sensor_tag_tr(device: Device[SensorTagService, T]) -> None:
    await _disable_sensor_tag(device)

# vim: sw=4:et:ai



#
# BTZen - library to asynchronously access Bluetooth devices.
#
# Copyright (C) 2015-2022 by Artur Wroblewski <wrobell@riseup.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from __future__ import annotations

import dataclasses as dtc
import typing as tp
from collections import defaultdict

from .data import T, AddressType, AnyTrigger, Converter, Make, NoTrigger, \
    ServiceType

ServiceRegistryDict: tp.TypeAlias = dict[
    'ServiceType',
    tuple['Service', Converter[T], AnyTrigger, 'AddressType']
]

# registry of known services
_SERVICE_REGISTRY = defaultdict['Make', ServiceRegistryDict[tp.Any]](ServiceRegistryDict[tp.Any])

@dtc.dataclass(frozen=True)
class Service:
    """
    Bluetooth service descriptor.

    :var uuid: UUID of Bluetooth service.
    """
    uuid: str

@dtc.dataclass(frozen=True)
class ServiceInterface(Service):
    """
    Bluetooth device information for device providing data via Bluez
    interface property.

    Example of interface for Battery Level Bluetooth characteristic is
    `org.bluez.Battery1` Bluez interface, which provides data via
    `Percentage` property.

    :var interface: Bluez interface name.
    :var property: Property name of the interface.
    :var type: Type of property value.
    """
    interface: str
    property: str
    type: str

@dtc.dataclass(frozen=True)
class ServiceCharacteristic(Service):
    """
    Bluetooth service descriptor for Bluetooth characteristic.

    :var uuid_data: UUID of characteristic to read data from.
    :var size: Length of data received from Bluetooth characteristic on
        read.
    """
    uuid_data: str
    size: int

def register_service(
        make: Make,
        service_type: ServiceType,
        service: Service,
        *,
        convert: Converter[T]=tp.cast(Converter[tp.Any], lambda v: v),
        trigger: AnyTrigger=NoTrigger(),
        address_type: AddressType=AddressType.PUBLIC,
    ) -> None:
    """
    Register service with data conversion function.
    """
    _SERVICE_REGISTRY[make][service_type] = (
        service, convert, trigger, address_type
    )

S = tp.TypeVar('S', bound=Service, covariant=True)

# vim: sw=4:et:ai



#
# BTZen - library to asynchronously access Bluetooth devices.
#
# Copyright (C) 2015-2022 by Artur Wroblewski <wrobell@riseup.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""
Operations for Bluetooth devices.

Basic operations

- `enable`
- `disable`
- `read`
- `write`
"""

from __future__ import annotations

import asyncio
import dataclasses as dtc
import inspect
import logging
import typing as tp
from collections.abc import AsyncIterator
from functools import singledispatch

from . import _btzen  # type: ignore
from .config import DEFAULT_DBUS_TIMEOUT
from .data import T
from .device import DeviceBase, Device, DeviceTrigger
from .service import Service, ServiceInterface, ServiceCharacteristic
from .session import get_session, connected, is_active

logger = logging.getLogger(__name__)

@singledispatch
async def read(device: DeviceBase[Service, T], *args: tp.Any) -> T:
    """
    Read data from Bluetooth device.

    The coroutine can raise cancellation error (`asyncio.CancelledError`),
    i.e. when the device is disconnected. The caller should handle the
    error, if it wants to continue reading data when the device is
    reconnected.

    :param device: Bluetooth device to read data from.
    :param args: Addition read arguments.
    """
    raise NotImplementedError(
        'Read function for {} is not implemented'.format(device)
    )

async def read_all(device: DeviceBase[Service, T]) -> AsyncIterator[T]:
    """
    Read all data from Bluetooth device.

    It ignores cancellation errors.

    :param device: Bluetooth device to read data from.
    """
    while is_active():
        try:
            value = await read(device)
        except asyncio.CancelledError as ex:
            # cancelled calls happen on device disconnection
            logger.info('{}: {}'.format(device, ex))
        else:
            yield value

@singledispatch
async def write(device: DeviceBase[Service, T], data: bytes) -> None:
    """
    Write data to Bluetooth device.

    :param device: Bluetooth device to write data to.
    :param data: Data to write to the Bluetooth device.
    """
    raise NotImplementedError(
        'Write function for {} is not implemented'.format(device)
    )

@singledispatch
async def enable(device: DeviceBase[Service, T]) -> None:
    """
    Enable and configure Bluetooth device.

    The function is called when a Bluetooth device is connected or
    reconnected.
    """

@singledispatch
async def disable(device: DeviceBase[Service, T]) -> None:
    """
    Disable Bluetooth device.

    The function is called when device is disconnected to release any
    resources held by the device.

    While the device might be gone and thus it is not possible to, for
    example, switch off a sensor on a Bluetooth device, there might be
    other resources to release like Bluez notification setup.
    """

@read.register
async def _read_dev(device: Device[ServiceCharacteristic, T]) -> T:
    async with connected(device) as session:
        bus = session.bus
        path = bus.characteristic_path(device.mac, device.service.uuid_data)
        assert path is not None

        task = session.create_future(
            device,
            _btzen.bt_read(bus.system_bus, path, DEFAULT_DBUS_TIMEOUT)
        )
        data = await task
        return device.convert(data)

@read.register
async def _read_dev_int(device: DeviceTrigger[ServiceInterface, T]) -> T:
    async with connected(device) as session:
        bus = session.bus
        srv = device.service

        task = session.create_future(
            device,
            bus._dev_property_get(device.mac, srv.property, srv.interface)
        )
        data = await task
        return device.convert(data)

@read.register
async def _read_dev_tr(device: DeviceTrigger[ServiceCharacteristic, T]) -> T:
    async with connected(device) as session:
        bus = session.bus
        path = bus.characteristic_path(device.mac, device.service.uuid_data)
        task = session.create_future(device, bus._gatt_get(path))
        data = await task
        return device.convert(data)

@enable.register
async def _enable_dev(device: Device[ServiceCharacteristic, T]) -> None:
    bus = get_session().bus
    await bus.ensure_characteristic_paths(device.mac, device.service.uuid_data)

@enable.register
async def _enable_int(device: DeviceTrigger[ServiceInterface, T]) -> None:
    bus = get_session().bus
    srv = device.service
    bus._dev_property_start(device.mac, srv.property, iface=srv.interface)

@enable.register
async def _enable_dev_trigger(
        device: DeviceTrigger[ServiceCharacteristic, T]
    ) -> None:

    bus = get_session().bus
    await bus.ensure_characteristic_paths(device.mac, device.service.uuid_data)

    path = bus.characteristic_path(device.mac, device.service.uuid_data)
    assert path is not None
    bus._gatt_start(path)
    logger.info('notifications enabled for {}'.format(path))

@disable.register
async def _disable_int(device: DeviceTrigger[ServiceInterface, T]) -> None:
    bus = get_session().bus
    srv = device.service
    bus._dev_property_stop(device.mac, srv.property, iface=srv.interface)

@disable.register
async def _disable_device_trigger(device: DeviceTrigger[ServiceCharacteristic, T]) -> None:
    bus = get_session().bus
    srv = device.service
    path = bus.characteristic_path(device.mac, srv.uuid_data)
    assert path is not None

    await disarm(
        'notifications for {} are disabled'.format(device),
        'cannot disable notifications for {}'.format(device),
        bus._gatt_stop, path
    )

async def write_config(mac: str, uuid: str, data: bytes) -> None:
    bus = get_session().bus
    path = bus.characteristic_path(mac, uuid)
    await _btzen.bt_write(
        bus.system_bus, path, data, DEFAULT_DBUS_TIMEOUT
    )

async def disarm(
        msg: str, warn: str, f: tp.Callable[..., tp.Any], *args: tp.Any
    ) -> None:
    try:
        if inspect.iscoroutinefunction(f):
            await f(*args)
        else:
            f(*args)
        logger.info(msg)
    except asyncio.CancelledError as ex:
        if not is_active():
            raise
        else:
            logger.warning(warn + ': ' + str(ex))
    except Exception as ex:
        logger.warning(warn + ': ' + str(ex))

# vim: sw=4:et:ai



#
# BTZen - library to asynchronously access Bluetooth devices.
#
# Copyright (C) 2015-2022 by Artur Wroblewski <wrobell@riseup.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import asyncio
import logging
import typing as tp
from collections.abc import Coroutine
from contextlib import asynccontextmanager
from contextvars import ContextVar

from .bus import Bus
from .error import BTZenError, CallError
from .data import T
from .device import DeviceBase
from .service import S

logger = logging.getLogger(__name__)

BT_SESSION = ContextVar['Session']('BT_SESSION')

class Session:
    """
    BTZen connection session.

    Await this object to close the session properly.
    """
    def __init__(self, bus: Bus):
        self.bus = bus
        self._is_active = False

        self._device_task: dict[DeviceBase[tp.Any, tp.Any], asyncio.Future[tp.Any]] = {}
        self._connection_task: dict[str, asyncio.Task[tp.Any]] = {}
        self._connection_status: dict[str, asyncio.Event] = {}

        self._event = asyncio.Event()

    def start(self) -> None:
        self._is_active = True

    def create_future(
        self, device: DeviceBase[S, T], f: Coroutine[None, None, T]
    ) -> asyncio.Future[T]:
        """
        Create future managed by BTZen connection session.

        If session stops, then all pending tasks are cancelled.
        """
        assert self._is_active

        task = asyncio.ensure_future(f)
        self._device_task[device] = task
        return task

    def add_connection_task(
        self, mac: str, f: Coroutine[None, None, T]
    ) -> asyncio.Task[T]:
        assert not self._is_active

        self._connection_status[mac] = asyncio.Event()

        task = asyncio.create_task(f)
        task.add_done_callback(self._stop)

        self._connection_task[mac] = task
        return task

    def set_connected(self, mac: str) -> None:
        assert self._is_active
        self._connection_status[mac].set()

    def set_disconnected(self, mac: str) -> None:
        self._connection_status[mac].clear()

    async def wait_connected(self, device: DeviceBase[S, T]) -> None:
        assert self._is_active

        event = self._connection_status.get(device.mac)
        if event is None:
            raise BTZenError(
                'Device with address {} not managed by BTZen connection manager'
                .format(device.mac)
            )
        # create future, so it can be cancelled when error happens in
        # connection management tasks
        task = self.create_future(device, event.wait())  # type: ignore
        await task

    def is_active(self) -> bool:
        return self._is_active

    def cancel_device_tasks(self, mac: str, msg: str) -> None:
        tasks = (t for d, t in self._device_task.items() if d.mac == mac)
        for t in tasks:
            t.cancel(msg=msg)

    def stop(self) -> None:
        self._is_active = False

        msg = 'BTZen session stopped'
        for t in self._device_task.values():
            t.cancel(msg=msg)
        for t in self._connection_task.values():
            t.cancel(msg=msg)

        logger.info('session is done')

    def _stop(self, task: asyncio.Task[T]) -> None:
        """
        Stop BTZen session if task is in error.
        """
        if task.done() and not task.cancelled() and task.exception():
            get_session().stop()
            try:
                task.result()
            except:
                logger.critical('Error in connection task', exc_info=True) 
                self._event.set()

    def __await__(self) -> tp.Generator[None, None, None]:
        # just wait forever and stop session on exit
        try:
            yield from self._event.wait().__await__()
        finally:
            self.stop()
        logger.info('session await exit')

@asynccontextmanager
async def connected(device: DeviceBase[S, T]) -> tp.AsyncIterator[Session]:
    session = get_session()

    if not session.is_active():
        raise asyncio.CancelledError(
            'BTZen is not running for device with address {}'
            .format(device.mac)
        )

    await session.wait_connected(device)
    if session.is_active():
        yield session

def get_session() -> Session:
    return BT_SESSION.get()

def is_active() -> bool:
    return get_session().is_active()

# vim: sw=4:et:ai



#
# BTZen - library to asynchronously access Bluetooth devices.
#
# Copyright (C) 2015-2022 by Artur Wroblewski <wrobell@riseup.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""
Bluetooth device descriptors, device object constructors and
modification functions.


                                   +--------------+
                                   | <<abstract>> |
                         +-------|>|  DeviceBase  |<|-------+
                         |         +--------------+         |
                         |                                  |
                         |                                  |
                     +---+----+   set_trigger() >    +---------------+
                     | Device |--------------------->| DeviceTrigger |
                     +--------+                      +---------------+

The following functions create Bluetooth device objects

- `accelerometer`
- `battery_level`
- `button`
- `humidity`
- `light`
- `light_rgb`
- `pressure`
- `serial`
- `temperature`
- `weight`

Depending on context a non-triggered or a triggered device is created. This
depends on a device itself and its make. If possible, non-triggered device
is created by default. Use `set_trigger` function to create tirggered
device.

It is not possible to create a non-triggered device from a triggered one.
"""

from __future__ import annotations

import dataclasses as dtc
import logging
import typing as tp
from functools import partial, singledispatch

from .data import T, AddressType, Converter, Make, NoTrigger, Trigger, \
    TriggerCondition, ServiceType, LightColor, Button, WeightData
from .service import S, Service, _SERVICE_REGISTRY

logger = logging.getLogger(__name__)

D = tp.TypeVar('D')

MakeNonTrigger = tp.Literal[Make.STANDARD, Make.SENSOR_TAG, Make.OSTC]

# key is tuple (base class, parameter class)
_PROXY_REGISTRY: dict[tuple[type, type], type] = {}

#
# define protocols for the constructors of devices objects
#
class Ctor(tp.Protocol[T]):
    """
    Constructor protocol for a device.

    By default it creates non-triggered device object. Exceptions are
    defined for devices, which work only with a trigger set.

    For example, by default a pressure device works without a trigger. But
    certain makes of pressure devices, i.e. Thingy:52, works only with
    a trigger set.
    """
    @tp.overload
    def __call__(
        self, mac: str, *, make: MakeNonTrigger=Make.STANDARD
    ) -> Device[Service, T]: ...

    @tp.overload
    def __call__(
        self, mac: str, *, make: tp.Literal[Make.THINGY52]
    ) -> DeviceTrigger[Service, T]: ...

    def __call__(
        self,
        mac: str,
        *,
        make: tp.Literal[Make.THINGY52] | MakeNonTrigger=Make.STANDARD
    ) -> Device[Service, T] | DeviceTrigger[Service, T]: ...

class CtorTrigger(tp.Protocol[T]):
    """
    Constructor for devices, which work only with triggers set.

    For example, any accelerometer device works with trigger set only.
    """
    def __call__(self, mac: str, make: Make=Make.STANDARD) -> DeviceTrigger[Service, T]: ...

#
# device objects
#
@dtc.dataclass(frozen=True)
class DeviceBase(tp.Generic[S, T]):
    """
    Abstract Bluetooth device descriptor.

    Implemenation notes

        BTZen handles complexity and diversity of non-compliant Bluetooth
        devices by using product of device and service types (`+` denotes
        that service class is upper bound for all other service classes)::

            (Device | DeviceTrigger, Service+)

        In Python, construction of product types can be achieved with
        `__new__` and `__class_getitem__` methods, for example::

            assert isinstance(Device(Service(...), ...), Device[Service])
            assert isinstance(Device(ServiceCharacteristic(...), ...), Device[ServiceCharacteristic])
            assert isinstance(DeviceTrigger(ServiceCharacteristic(...), ...), DeviceTrigger[ServiceCharacteristic])

        For BTZen purposes, instances of such types have to be dispatched
        with Python single-dispatch generic functions. The functions
        implement behaviour for the Bluetooth devices, i.e.::

            @read.register
            async def _read_sensor_tag(device: Device[SensorTagService, T]):
                ...

        where `SensorTagService` is an example of service class, which
        describes Texas Instrument Sensor Tag Bluetooth device sensors.

        The `__class_getitem__` method creates proxy classes for the device
        and service product types plus Python generic psuedo-types. This
        allows to use them with single-dispatch generic functions in
        Python.

        The `__new__` method creates instance using the proxy class.

        Why "pseudo-type"? The types defined in Python typing module (as of
        Python 3.9) are not real Python types, for example all statements
        below return false or raise error::

            issubclass(Generic, type)  # False
            issubclass(Generic[TypeVar('T')], type)  # type error "arg1 must be a class"

        As such, the pseudo-types cannot be used by single-dispatch generic
        functions in Python at the moment. See also

            https://bugs.python.org/issue34498

        Single-dispatch generic functions, which implement behaviour for
        the product types, can be listed with script
        `scripts/btzen-service-impl`.

    :var service: Bluetooth service descriptor.
    :var mac: MAC address of Bluetooth device.
    :var address_type: Bluetooth device address type.
    :var convert: Function to convert binary data provided by service
        to a value.
    """
    service: S
    mac: str
    address_type: AddressType
    convert: Converter[T]

    def __new__(cls, *args, **kw):  # type: ignore
        tv = type(args[0])
        return cls[tv, T](*args, **kw)  # type: ignore

    def __class_getitem__(cls, cls_param: tuple[type[S], tp.TypeVar]) -> type[Device[S, T]]:
        ok = (
            isinstance(cls_param, tuple)
            and len(cls_param) == 2
            and (dtc.is_dataclass(cls_param[0]) or cls_param[0] == S) # type: ignore
            and (
                isinstance(cls_param[1], (tp.TypeVar, type))
                or cls_param[1] is tp.Any
            )
        )

        if not ok:
            raise TypeError(
                'Class subscript requires dataclass and type variable. Got {}'
                .format(cls_param)
            )

        cls_pt = cls_param[0]
        is_s = cls_pt == S  # type: ignore

        # this allows use of subclasses
        if is_s:
            return cls  # type: ignore

        key = cls, cls_pt
        t = _PROXY_REGISTRY.get(key)
        if t is None:
            # find base class of new dataclass, so subclassing works
            get_parent = cls.__class_getitem__
            mro = [] if is_s else cls_pt.mro()[1:]
            items = (c for c in mro if issubclass(c, S.__bound__))  # type: ignore
            base = next(items, None)
            bases = (get_parent((base, cls_param[1])), object,) if base else (object,)

            # create new dataclass based on main class `cls` and the
            # parameter class `cls_pt`
            t = type('{}[{}]'.format(cls.__name__, cls_pt.__name__), bases, {})
            fields = cls.__dataclass_fields__ 
            t = dtc.make_dataclass(t.__name__, fields, bases=bases, frozen=True)
            t.__product__ = (cls, cls_pt)  # type: ignore
            _PROXY_REGISTRY[key] = t
        return t

@dtc.dataclass(frozen=True)
class Device(DeviceBase[S, T]):
    """
    Bluetooth device descriptor.
    """

@dtc.dataclass(frozen=True)
class DeviceTrigger(DeviceBase[S, T]):
    """
    Bluetooth device descriptor with trigger.

    :var trigger: Trigger information.
    """
    trigger: Trigger

def create_device(
        service: Service,
        mac: str,
        *,
        address_type: AddressType=AddressType.PUBLIC,
        convert: Converter[T]=tp.cast(Converter[tp.Any], lambda v: v),
    ) -> Device[Service, T]:
    """
    Create Bluetooth device for a Bluetooth service.
    """
    return Device(service, mac, address_type=address_type, convert=convert)

def set_interval(
        device: DeviceBase[S, T],
        interval: float,
    ) -> DeviceTrigger[S, T]:
    """
    Set fixed time interval for Bluetooth Environmental Sensing device.

    This is equivalent to::

        set_trigger(TriggerCondition.FIXED_TIME, interval)

    :param device: Bluetooth device descriptor.
    :param interval: Interval in seconds.
    """
    return set_trigger(device, TriggerCondition.FIXED_TIME, operand=interval)

@singledispatch
def set_trigger(
        device: DeviceBase[S, T],
        condition: TriggerCondition,
        *,
        operand: tp.Optional[float]=None,
        ) -> DeviceTrigger[S, T]:
    """
    Set trigger for Bluetooth Environmental Sensing device.
    """
    return DeviceTrigger(
        device.service,
        device.mac,
        device.address_type,
        device.convert,
        Trigger(condition, operand),
    )

def set_address_type(device: D, address_type: AddressType) -> D:
    """
    Set connection address type for a Bluetooth device.

    :param device: Bluetooth device descriptor.
    :param address_type: Bluetooth device connection address type.
    """
    return dtc.replace(device, address_type=address_type)  # type: ignore

def _create_device(
        service_type: ServiceType,
        mac: str,
        *,
        make: Make=Make.STANDARD,
    ) -> DeviceBase[Service, tp.Any]:

    srv, conv, trigger, addr_type = _SERVICE_REGISTRY[make][service_type]
    dev = create_device(srv, mac, address_type=addr_type, convert=conv)

    if isinstance(trigger, NoTrigger):
        return dev
    else:
        return set_trigger(dev, trigger.condition, operand=trigger.operand)

accelerometer = tp.cast(
    CtorTrigger[tuple[float, float, float]],
    partial(_create_device, ServiceType.ACCELEROMETER),
)

battery_level = tp.cast(CtorTrigger[int], partial(_create_device, ServiceType.BATTERY_LEVEL))
battery_level.__doc__ = """
The current charge level of a Bluetooth device battery.

:param mac: MAC address of Bluetooth device.
:param make: Bluetooth device make.

.. seealso:: `Make`
"""

button = tp.cast(CtorTrigger[Button], partial(_create_device, ServiceType.BUTTON))
humidity = tp.cast(Ctor[float], partial(_create_device, ServiceType.HUMIDITY))
light = tp.cast(Ctor[float], partial(_create_device, ServiceType.LIGHT))
light_rgb = tp.cast(Ctor[LightColor], partial(_create_device, ServiceType.LIGHT_RGB))
pressure = tp.cast(Ctor[float], partial(_create_device, ServiceType.PRESSURE))
serial = tp.cast(Ctor[bytes], partial(_create_device, ServiceType.SERIAL))
temperature = tp.cast(Ctor[float], partial(_create_device, ServiceType.TEMPERATURE))
weight = tp.cast(CtorTrigger[WeightData], partial(_create_device, ServiceType.WEIGHT_MEASUREMENT))

# vim: sw=4:et:ai



#
# BTZen - library to asynchronously access Bluetooth devices.
#
# Copyright (C) 2015-2022 by Artur Wroblewski <wrobell@riseup.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""
Bluetooth Weight service implementation.
"""

import dataclasses as dtc
import enum
import struct
import typing as tp

from .data import T, Make, ServiceType, Trigger, TriggerCondition, \
    WeightData, WeightFlags
from .service import register_service, ServiceCharacteristic
from .util import to_uuid

@dtc.dataclass(frozen=True)
class MiScaleWeightData(WeightData):
    """
    Weight measurement data for Mi Smart Scale.

    :var stabilized: Indicates if weight stabilized.
    :var load_removed: Indicates if load is removed.
    """
    stabilized: tp.Optional[bool] = True
    load_removed: tp.Optional[bool] = False

def convert_weight(data: bytes) -> MiScaleWeightData:
    flags, weight = struct.unpack('<BH', data[0:3])
    flags = WeightFlags(flags)

    stabilized = bool(flags & WeightFlags.RESERVED_2)
    load_removed = bool(flags & WeightFlags.RESERVED_4)

    return MiScaleWeightData(
        flags,
        weight * 0.005,
        stabilized=stabilized,
        load_removed=load_removed
    )

register_service(
    Make.MI_SMART_SCALE,
    ServiceType.WEIGHT_MEASUREMENT,
    ServiceCharacteristic(
        to_uuid(0x181d),
        to_uuid(0x2a9d),
        9
    ),
    convert=convert_weight,
    trigger=Trigger(TriggerCondition.ON_CHANGE),
)

# vim: sw=4:et:ai



#!/usr/bin/env python3
#
# BTZen - library to asynchronously access Bluetooth devices.
#
# Copyright (C) 2015-2022 by Artur Wroblewski <wrobell@riseup.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""
Build setup for the BTZen library.
"""

from setuptools import setup, Extension
from Cython.Build import cythonize

setup(
    ext_modules=cythonize([
        Extension('btzen._sd_bus', ['btzen/_sd_bus.pyx'], libraries=['systemd']),
        Extension('btzen._btzen', ['btzen/_btzen.pyx'], libraries=['systemd']),
        Extension('btzen._cm', ['btzen/_cm.pyx'], libraries=['systemd']),
    ])
)

# vim: sw=4:et:ai
